package ru.dbotthepony.kstarbound.io

import it.unimi.dsi.fastutil.bytes.ByteArrayList
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kvector.api.IStruct2d
import ru.dbotthepony.kvector.api.IStruct2i
import ru.dbotthepony.kvector.vector.Vector2d
import ru.dbotthepony.kvector.vector.Vector2i
import java.io.DataInput
import java.io.DataOutput
import java.io.IOException
import java.io.InputStream
import java.io.OutputStream
import java.io.RandomAccessFile
import java.math.BigDecimal
import java.util.*
import java.util.function.IntConsumer
import java.util.function.IntSupplier
import kotlin.collections.ArrayList

data class VarIntReadResult(val value: Int, val cells: Int)
data class VarLongReadResult(val value: Long, val cells: Int)

private fun readVarLongInfo(supplier: IntSupplier): VarLongReadResult {
	var result = 0L
	var read = supplier.asInt
	var i = 1

	while (true) {
		result = (result shl 7) or (read.toLong() and 0x7F)

		if (read and 0x80 == 0)
			break

		read = supplier.asInt
		i++
	}

	return VarLongReadResult(result, i)
}

private fun readVarIntInfo(supplier: IntSupplier): VarIntReadResult {
	val read = readVarLongInfo(supplier)
	return VarIntReadResult(read.value.toInt(), read.cells)
}

private fun readVarInt(supplier: IntSupplier): Int {
	return readVarIntInfo(supplier).value
}

private fun readVarLong(supplier: IntSupplier): Long {
	return readVarLongInfo(supplier).value
}

private fun Int.fromSignedVar(): Int {
	val sign = this and 1

	if (sign == 0)
		return this ushr 1
	else
		return -(this ushr 1) - 1
}

private fun VarIntReadResult.fromSignedVar(): VarIntReadResult {
	val sign = this.value and 1

	if (sign == 0)
		return VarIntReadResult(this.value ushr 1, this.cells)
	else
		return VarIntReadResult(-(this.value ushr 1) - 1, this.cells)
}

private fun Long.fromSignedVar(): Long {
	val sign = this and 1

	if (sign == 0L)
		return this ushr 1
	else
		return -(this ushr 1) - 1L
}

private fun VarLongReadResult.fromSignedVar(): VarLongReadResult {
	val sign = this.value and 1

	if (sign == 0L)
		return VarLongReadResult(this.value ushr 1, this.cells)
	else
		return VarLongReadResult(-(this.value ushr 1) - 1, this.cells)
}

fun RandomAccessFile.readVarLong() = readVarLong(::read)
fun RandomAccessFile.readVarInt() = readVarInt(::read)
fun RandomAccessFile.readVarLongInfo() = readVarLongInfo(::read)
fun RandomAccessFile.readVarIntInfo() = readVarIntInfo(::read)
fun InputStream.readVarLong() = readVarLong(::read)
fun InputStream.readVarInt() = readVarInt(::read)
fun InputStream.readVarLongInfo() = readVarLongInfo(::read)
fun InputStream.readVarIntInfo() = readVarIntInfo(::read)

fun RandomAccessFile.readSignedVarLong() = readVarLong(::read).fromSignedVar()
fun RandomAccessFile.readSignedVarInt() = readVarInt(::read).fromSignedVar()
fun RandomAccessFile.readSignedVarLongInfo() = readVarLongInfo(::read).fromSignedVar()
fun RandomAccessFile.readSignedVarIntInfo() = readVarIntInfo(::read).fromSignedVar()
fun InputStream.readSignedVarLong() = readVarLong(::read).fromSignedVar()
fun InputStream.readSignedVarInt() = readVarInt(::read).fromSignedVar()
fun InputStream.readSignedVarLongInfo() = readVarLongInfo(::read).fromSignedVar()
fun InputStream.readSignedVarIntInfo() = readVarIntInfo(::read).fromSignedVar()

private fun writeVarLong(write: IntConsumer, value: Long): Int {
	var value = value
	var i = 0

	do {
		val toWrite = (value and 0x7F).toInt()
		value = value ushr 7

		if (value == 0L)
			write.accept(toWrite)
		else
			write.accept(toWrite or 0x80)

		i++
	} while (value != 0L)

	return i
}

private fun writeVarInt(write: IntConsumer, value: Int): Int {
	return writeVarLong(write, value.toLong())
}

private fun Int.toSignedVar(): Int {
	if (this >= 0)
		return this shl 1
	else
		return (this shl 1) or 1
}

private fun Long.toSignedVar(): Long {
	if (this >= 0)
		return this shl 1
	else
		return ((-this - 1) shl 1) or 1L
}

fun RandomAccessFile.writeVarLong(value: Long) = writeVarLong(::write, value)
fun RandomAccessFile.writeVarInt(value: Int) = writeVarInt(::write, value)
fun OutputStream.writeVarLong(value: Long) = writeVarLong(::write, value)
fun OutputStream.writeVarInt(value: Int) = writeVarInt(::write, value)

fun RandomAccessFile.writeSignedVarLong(value: Long) = writeVarLong(::write, value.toSignedVar())
fun RandomAccessFile.writeSignedVarInt(value: Int) = writeVarInt(::write, value.toSignedVar())
fun OutputStream.writeSignedVarLong(value: Long) = writeVarLong(::write, value.toSignedVar())
fun OutputStream.writeSignedVarInt(value: Int) = writeVarInt(::write, value.toSignedVar())

fun RandomAccessFile.readString(length: Int): String {
	require(length >= 0) { "Invalid length $length" }

	val bytes = ByteArray(length)
	try {
		readFully(bytes)
	} catch(err: Throwable) {
		throw IOException("Tried to read string with length of $length", err)
	}

	return bytes.toString(Charsets.UTF_8)
}

fun InputStream.readString(length: Int, allowSmaller: Boolean = false): String {
	require(length >= 0) { "Invalid length $length" }

	val bytes = ByteArray(length)

	try {
		val read = readNBytes(bytes, 0, length)

		if (!allowSmaller)
			require(read == bytes.size) { "Read $read bytes, expected ${bytes.size}" }
	} catch(err: Throwable) {
		throw IOException("Tried to read string with length of $length", err)
	}

	return bytes.toString(Charsets.UTF_8)
}

fun InputStream.readByteChar(): Char {
	return read().toChar()
}

fun InputStream.readHeader(header: String) {
	for ((i, char) in header.withIndex()) {
		val read = readByteChar()

		if (read != char) {
			throw IllegalArgumentException("Malformed header at byte $i: expected $char (${char.code}) got $read (${char.code})")
		}
	}
}

fun OutputStream.writeUTF(value: String) {
	write(value.toByteArray().also { check(!it.any { it.toInt() == 0 }) { "Provided UTF string contains NUL" } })
	write(0)
}

fun InputStream.readUTF(): String {
	val bytes = ByteArrayList()
	var read = read()

	while (read != 0) {
		bytes.add(read.toByte())
		read = read()
	}

	return String(bytes.toByteArray())
}

fun InputStream.readUUID(): UUID {
	return UUID(readLong(), readLong())
}

fun OutputStream.writeUUID(value: UUID) {
	writeLong(value.mostSignificantBits)
	writeLong(value.leastSignificantBits)
}

fun OutputStream.writeVec2i(value: IStruct2i) {
	writeSignedVarInt(value.component1())
	writeSignedVarInt(value.component2())
}

fun OutputStream.writeVec2d(value: IStruct2d) {
	writeDouble(value.component1())
	writeDouble(value.component2())
}

fun InputStream.readVec2i(): Vector2i {
	return Vector2i(readSignedVarInt(), readSignedVarInt())
}

fun InputStream.readVec2d(): Vector2d {
	return Vector2d(readDouble(), readDouble())
}

fun InputStream.readChunkPos(): ChunkPos {
	return ChunkPos(readSignedVarInt(), readSignedVarInt())
}

fun <S : OutputStream, V> S.writeCollection(collection: Collection<V>, writer: S.(V) -> Unit) {
	writeVarInt(collection.size)

	for (value in collection) {
		writer(this, value)
	}
}

fun <S : InputStream, V, C : MutableCollection<V>> S.readCollection(reader: S.() -> V, factory: (Int) -> C): C {
	val size = readVarInt()
	val collection = factory.invoke(size)

	for (i in 0 until size) {
		collection.add(reader(this))
	}

	return collection
}

fun <S : InputStream, V> S.readCollection(reader: S.() -> V) = readCollection(reader, ::ArrayList)

fun OutputStream.writeInt(value: Int) {
	if (this is DataOutput) {
		writeInt(value)
		return
	}

	write(value ushr 24)
	write(value ushr 16)
	write(value ushr 8)
	write(value)
}

fun InputStream.readInt(): Int {
	if (this is DataInput) {
		return readInt()
	}

	return (read() shl 24) or (read() shl 16) or (read() shl 8) or read()
}

fun OutputStream.writeLong(value: Long) {
	if (this is DataOutput) {
		writeLong(value)
		return
	}

	write((value ushr 48).toInt())
	write((value ushr 40).toInt())
	write((value ushr 32).toInt())
	write((value ushr 24).toInt())
	write((value ushr 16).toInt())
	write((value ushr 8).toInt())
	write(value.toInt())
}

fun InputStream.readLong(): Long {
	if (this is DataInput) {
		return readLong()
	}

	return (read().toLong() shl 48) or
		(read().toLong() shl 40) or
		(read().toLong() shl 32) or
		(read().toLong() shl 24) or
		(read().toLong() shl 16) or
		(read().toLong() shl 8) or
		read().toLong()
}

fun OutputStream.writeFloat(value: Float) = writeInt(value.toBits())
fun InputStream.readFloat() = Float.fromBits(readInt())
fun OutputStream.writeDouble(value: Double) = writeLong(value.toBits())
fun InputStream.readDouble() = Double.fromBits(readLong())

fun InputStream.readBinaryString(): String {
	val size = readVarInt()
	require(size >= 0) { "Negative payload size: $size" }
	val bytes = ByteArray(size)
	read(bytes)
	return bytes.decodeToString()
}

fun OutputStream.writeBinaryString(input: String) {
	val bytes = input.encodeToByteArray()
	writeVarInt(bytes.size)
	write(bytes)
}
